/** 
 * XMLSec library
 *
 * Keys
 *
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>
 
#include <libxml/tree.h>
#include <openssl/x509.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/x509.h>

#include <xmlsec/hmac.h>
#include <xmlsec/dsa.h>
#include <xmlsec/rsa.h>

#include <xmlsec/keys.h>

static xmlSecKeyId xmlSecAllKeyIds[100];

static xmlChar* 	xmlSecKeyNameRead		(const xmlNodePtr keyNameNode);
static int 		xmlSecKeyNameWrite		(xmlNodePtr keyNameNode,
							 xmlSecKeyPtr key,
							 xmlSecKeyType type);
static xmlSecKeyPtr	xmlSecKeyValueRead		(const xmlNodePtr keyValueNode, 
							 xmlSecTransformId id, 
							 xmlSecKeyType type);
static int 		xmlSecKeyValueWrite		(xmlNodePtr keyValueNode,
							 xmlSecKeyPtr key,
							 xmlSecKeyType type);
static xmlSecKeyPtr	xmlSecRetrievalMethodRead	(const xmlNodePtr retrievalMethodNode,
							 xmlSecTransformId id,
							 xmlSecKeyType type,
							 int level,
							 const xmlSecKeysReadContext *ctx);
static xmlSecKeyPtr	xmlSecKeyNodesListRead		(xmlNodePtr cur, 
							 xmlSecTransformId id, 
							 xmlSecKeyType type, 
							 int level, 
							 const xmlSecKeysReadContext *ctx);

/**
 * xmlSecKeyInit
 * 
 * Initializes the keys list
 */
void		
xmlSecKeysInit(void) {
    int i = 0;
    
    xmlSecAllKeyIds[i++] = xmlSecMacHmacKey;
    xmlSecAllKeyIds[i++] = xmlSecSignDsaKey;
    xmlSecAllKeyIds[i++] = xmlSecSignRsaKey;

    /* MUST be the last in the list */  
    xmlSecAllKeyIds[i++] = xmlSecKeyIdUnknown;
}

/**
 * xmlSecKeySearcCtxFindKey:
 * @keysContext:
 * @id:
 * @type:
 * @name:
 *
 */
xmlSecKeyPtr 	
xmlSecKeySearcCtxFindKey(const xmlSecKeysReadContext *keysContext,
			 xmlSecTransformId id, xmlSecKeyType type,
			 const xmlChar *name) {
    static const char func[] = "xmlSecKeySearcCtxFindKey";
    xmlSecKeyPtr key = NULL;
    
    if(keysContext == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key search context is null\n",
	    func);	
#endif
	return(NULL);	
    }
    
    if(keysContext->findKeyCallback != NULL) {
	key = (keysContext->findKeyCallback)(keysContext->findKeyContext,
					      id, type, name);
    }
    return(key);
}

/**
 * xmlSecKeyCreate
 * @id:
 * @origin:
 *
 * Creates new key (wrapper for xmlSecKey::create method)
 *
 */
xmlSecKeyPtr	
xmlSecKeyCreate(xmlSecKeyId id, xmlSecKeyOrigin origin)  {
    static const char func[] = "xmlSecKeyCreate";
    xmlSecKeyPtr key;
    
    if((id == xmlSecKeyIdUnknown) || (id->create == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: id or create method is not defined\n",
	    func);	
#endif
	return(NULL);	
    }
    
    key = id->create(id);
    if(key == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key creation failed\n");	
#endif
	return(NULL);	
    }
    key->origin = origin;
    return(key);
}

/**
 * xmlSecKeyDestroy
 * @key:
 * @forceDestroy:
 *
 * Destroys the key. If the key origin has a KeyManager marker 
 * the key is *not* destroyed unless @forceDestroy is not zero
 */
void
xmlSecKeyDestroy(xmlSecKeyPtr key,  int forceDestroy) {
    static const char func[] = "xmlSecKeyDestroy";
    
    if((!xmlSecKeyIsValid(key)) || (key->id->destroy == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or destroy method is not defined\n",
	    func);	
#endif
	return;
    }
    
    if((key->origin & xmlSecKeyOriginKeyManager) && (!forceDestroy)) {
	/* 
	 * we do not destroy keys from key manager unless it is
	 * specially requested
	 */
	return;
    }
    if(key->name != NULL) {
	xmlFree(key->name); 
	key->name = NULL;
    }
    if(key->x509Chain != NULL) {	
	sk_X509_pop_free(key->x509Chain, X509_free); 
	key->x509Chain = NULL;
    }
    key->id->destroy(key);
}

/**
 * xmlSecKeyGenerate
 * @id:
 * @data:
 *
 * Generates the key (wrapper for xmlSecKey::generate method)
 */
xmlSecKeyPtr	
xmlSecKeyGenerate(xmlSecKeyId id, void *data) {
    static const char func[] = "xmlSecKeyGenerate";
    xmlSecKeyPtr key;
    int ret;
    
    if((id == NULL) || (id->generate == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: id is invalid or generate method is not defined\n",
	    func);	
#endif
	return(NULL);
    }
    
    key = xmlSecKeyCreate(id, xmlSecKeyOriginDefault);
    if(key == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key\n",
	    func);	
#endif
	return(NULL);    
    }
    
    ret = (id->generate)(key, data);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to generate key\n",
	    func);	
#endif
	xmlSecKeyDestroy(key, 1);
	return(NULL);    
    }
    
    return (key);
}

/**
 * xmlSecKeyReadXml
 * @node:
 * 
 * Reads the key form XML doc
 */
xmlSecKeyPtr	
xmlSecKeyReadXml(xmlSecKeyId id, xmlNodePtr node) {
    static const char func[] = "xmlSecKeyReadXml";
    xmlSecKeyPtr key;
    int ret;
    
    if(node == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: node is null\n",
	    func);	
#endif
	return(NULL);
    }

    key = xmlSecKeyCreate(id, xmlSecKeyOriginDefault);
    if(key == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key\n",
	    func);	
#endif
	return(NULL);    
    }
    
    ret = (id->read)(key, node);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read key\n",
	    func);	
#endif
	xmlSecKeyDestroy(key, 1);
	return(NULL);    
    }
    
    return(key);
}

/**
 * xmlSecKeyWriteXml
 * @key:
 * @type:
 * @parent:
 * 
 * Writes the key into XML node and adds it to the list of @parent children
 * (if parent is not null).
 */
int
xmlSecKeyWriteXml(xmlSecKeyPtr key, xmlSecKeyType type, xmlNodePtr node) {
    static const char func[] = "xmlSecKeyWriteXml";
    int ret;
    
    if((!xmlSecKeyIsValid(key)) || (node == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or parent is null\n",
	    func);	
#endif
	return(-1);
    }
    
    /* write key */
    ret = key->id->write(key, type, node);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key\n",
	    func);	
#endif
	return(-1);	    
    }
    return(0);    
}

/** 
 * XMLDsig:KeyInfo methods
 * http://www.w3.org/TR/xmlSec-core/#sec-KeyInfo
 *
 * The KeyInfo Element
 *
 * KeyInfo is an optional element that enables the recipient(s) to obtain 
 * the key needed to validate the signature.  KeyInfo may contain keys, 
 * names, certificates and other public key management information, such as 
 * in-band key distribution or key agreement data. 
 * 
 *  Schema Definition:
 *
 *  <element name="KeyInfo" type="ds:KeyInfoType"/> 
 *  <complexType name="KeyInfoType" mixed="true">
 *    <choice maxOccurs="unbounded"> 
 *       <element ref="ds:KeyName"/> 
 *       <element ref="ds:KeyValue"/> 
 *       <element ref="ds:RetrievalMethod"/> 
 *       <element ref="ds:X509Data"/> 
 *       <element ref="ds:PGPData"/> 
 *       <element ref="ds:SPKIData"/>
 *       <element ref="ds:MgmtData"/>
 *       <any processContents="lax" namespace="##other"/>
 *       <!-- (1,1) elements from (0,unbounded) namespaces -->
 *    </choice>
 *    <attribute name="Id" type="ID" use="optional"/>
 *  </complexType>
 *    
 * DTD:
 *    
 * <!ELEMENT KeyInfo (#PCDATA|KeyName|KeyValue|RetrievalMethod|
 *                    X509Data|PGPData|SPKIData|MgmtData %KeyInfo.ANY;)* >      
 * <!ATTLIST KeyInfo  Id  ID   #IMPLIED >
 *  
 */ 
/**
 * xmlSecKeyInfoRead:
 * @keyInfoNode:
 * @id:			the trasnform we need key for
 * @type:		the type of the key we are looking for
 * @ctx:
 *
 * Reads the key from key info node. We ignore the unknown nodes and hope 
 * for the best
 */
xmlSecKeyPtr	
xmlSecKeyInfoRead(const xmlNodePtr keyInfoNode, xmlSecTransformId id,
	    	xmlSecKeyType type, const xmlSecKeysReadContext *ctx) {
    static const char func[] = "xmlSecKeyInfoRead";
    xmlSecKeyPtr key = NULL;
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL) || (ctx == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyInfo node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }
    cur = xmlSecGetNextElementNode(keyInfoNode->children); 
    if(keyInfoNode->children != NULL) {
	key = xmlSecKeyNodesListRead(cur, id, type, 0, ctx);
    }
    return(key);
}

/**
 * xmlSecKeyNodesListRead:
 * @cur:
 * @id:			the trasnform we need key for
 * @type:		the type of the key we are looking for
 * @level:
 * @ctx:
 *
 */
static xmlSecKeyPtr
xmlSecKeyNodesListRead(xmlNodePtr cur, xmlSecTransformId id, xmlSecKeyType type, 
		int level, const xmlSecKeysReadContext *ctx) { 
    static const char func[] = "xmlSecKeyNodesListRead";
    xmlChar *keyName;
    xmlSecKeyPtr key;
    
    if((cur == NULL) || (ctx == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyInfo node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    key = NULL;
    keyName = NULL;
    while ((key == NULL) && (cur != NULL)) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "KeyName", xmlSecDSigNs)) {
	    if(keyName != NULL) {
	        xmlFree(keyName);
	    }		
	    keyName = xmlSecKeyNameRead(cur);
	    if((keyName != NULL) && (ctx->allowedOrigins & xmlSecKeyOriginKeyManager) && 
			(ctx->allowedOrigins & xmlSecKeyOriginKeyName)) {
		key = xmlSecKeySearcCtxFindKey(ctx, id, type, keyName);
		if(key != NULL) {
		    return(key);
		}
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "KeyValue", xmlSecDSigNs) &&
			(ctx->allowedOrigins & xmlSecKeyOriginKeyValue)) {
	    key = xmlSecKeyValueRead(cur, id, type);
	    if((key != NULL) && (keyName != NULL)){
	        key->name = keyName;
	        keyName = NULL;
	        return(key);
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "RetrievalMethod", xmlSecDSigNs) &&
			(ctx->allowedOrigins & (xmlSecKeyOriginRetrievalDocument |
						xmlSecKeyOriginRetrievalRemote)) &&
			(level < ctx->maxRetrievals)) {
	    key = xmlSecRetrievalMethodRead(cur, id, type, level, ctx);
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509Data", xmlSecDSigNs) && 
			(ctx->allowedOrigins & xmlSecKeyOriginX509)) {
	    key = xmlSecX509DataNodeRead(cur, id, type, ctx);
	}
	
	/* TODO: add more nodes (pgp, spki, etc) */
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    if(keyName != NULL) {
	xmlFree(keyName);
    }
    return(key);
}

/**
 * xmlSecKeyInfoWrite
 * @keyInfoNode:
 * @key:
 * @type:
 *
 * Writes the key in the key info node according to the KeyInfo node
 * content
 */
int
xmlSecKeyInfoWrite(xmlNodePtr keyInfoNode, const xmlSecKeyPtr key,
		    xmlSecKeyType type) {
    static const char func[] = "xmlSecKeyInfoWrite";
    xmlNodePtr cur;
    int ret;
        
    if(keyInfoNode == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfo node  is null\n",
	    func);	
#endif
	return(-1);
    }

    ret = 0;
    cur = xmlSecGetNextElementNode(keyInfoNode->children);
    while(cur != NULL) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "KeyName", xmlSecDSigNs)) {
	    ret = xmlSecKeyNameWrite(cur, key, type);
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "KeyValue", xmlSecDSigNs)) {
	    ret = xmlSecKeyValueWrite(cur, key, type);
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509Data", xmlSecDSigNs)) {
	    ret = xmlSecX509DataNodeWrite(cur, key); 
	}
	/* TODO: add retrieval method, etc. */
	if(ret < 0) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: keyInfo failed for \"%s\"\n",
		func, cur->name);	
#endif
	    return(-1);	    
	}
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    return(0);
}		    


/**
 * xmlSecKeyNameRead:
 * @keyNameNode:
 *
 *
 */
static xmlChar*
xmlSecKeyNameRead(const xmlNodePtr keyNameNode) {
    static const char func[] = "xmlSecKeyNameRead";    
    xmlChar *content;
    
    if(keyNameNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyName node is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    content = xmlNodeGetContent(keyNameNode);
    /* TODO: decode key name */    
    return(content);
}

/** 
 * xmlSecKeyNameWrite:
 * @keyNameNode:
 * @key:
 * @type:
 *
 *
 */
static int 
xmlSecKeyNameWrite(xmlNodePtr keyNameNode, xmlSecKeyPtr key, xmlSecKeyType type) {
    static const char func[] = "xmlSecKeyNameWrite";
    
    if((keyNameNode == NULL) || (!xmlSecKeyIsValid(key))) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyName node or key is not valid\n",
	    func);	
#endif
	return(-1);    
    }
    
    if(key->name != NULL) {
	/* TODO: encode the key name! */
	xmlNodeSetContent(keyNameNode, key->name);
    }
    return(0);
}


/**
 * xmlSecKeyValueRead
 * @keyValueNode:
 * @id:
 * @type:
 *
 * http://www.w3.org/TR/xmldsig-core/#sec-KeyValue
 *
 * The KeyValue element contains a single public key that may be useful in 
 * validating the signature. Structured formats for defining DSA (REQUIRED) 
 * and RSA (RECOMMENDED) public keys are defined in Signature Algorithms 
 * (section 6.4). The KeyValue element may include externally defined 
 * public keys values represented as PCDATA or element types from an external 
 * namespace.
 *
 * Schema Definition:
 *
 * <element name="KeyValue" type="ds:KeyValueType"/> 
 *   <complexType name="KeyValueType" mixed="true">
 *     <choice>
 *       <element ref="ds:DSAKeyValue"/>
 *       <element ref="ds:RSAKeyValue"/>
 *       <any namespace="##other" processContents="lax"/>
 *     </choice>
 *   </complexType>
 *    
 * DTD:
 *  
 * <!ELEMENT KeyValue (#PCDATA|DSAKeyValue|RSAKeyValue %KeyValue.ANY;)* >
 *       
 * =========================================================================
 * Support for private keys is added (@type parameter)
 */
static xmlSecKeyPtr
xmlSecKeyValueRead(const xmlNodePtr keyValueNode, xmlSecTransformId id, 
		xmlSecKeyType type) {
    static const char func[] = "xmlSecKeyValueRead";   
    xmlNodePtr cur; 
    xmlSecKeyId keyId;
    xmlSecKeyPtr key;
    size_t i;
    
    if(keyValueNode == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyValue node is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    key = NULL;
    cur = xmlSecGetNextElementNode(keyValueNode->children);    
    while(cur != NULL) {
        for(i = 0; xmlSecAllKeyIds[i] != xmlSecKeyIdUnknown; ++i) {
	    keyId = xmlSecAllKeyIds[i];
	    if((id != xmlSecTransformUnknown) && 
	        !xmlSecBinTransformIdCheckKeyId(id, keyId)) continue;
	    if(xmlSecCheckNodeName(cur, keyId->keyValueNodeName, 
					keyId->keyValueNodeNs)) {
		key = xmlSecKeyReadXml(keyId, cur);
		break;
	    }
	}
	if(key != NULL) {
	    if((key->type == type) || (type == xmlSecKeyTypeAny)) {
		return(key);
	    } else {
		xmlSecKeyDestroy(key, 0);
		key = NULL;
	    }
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }
    return(key);
}

/**
 * xmlSecKeyValueWrite:
 * @keyValueNode:
 * @key:
 * @type:
 *
 *
 */
static int 
xmlSecKeyValueWrite(xmlNodePtr keyValueNode, xmlSecKeyPtr key,  xmlSecKeyType type) {
    static const char func[] = "xmlSecKeyValueWrite";
    xmlNodePtr cur;
    int ret;
    
    if((keyValueNode == NULL) || (!xmlSecKeyIsValid(key))) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyValue node or key is null\n",
	    func);	
#endif
	return(-1);
    }
    
    /* remove all existing key value */
    xmlNodeSetContent(keyValueNode, NULL);
    
    /* create key node */
    cur = xmlSecAddChild(keyValueNode, key->id->keyValueNodeName, key->id->keyValueNodeNs);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key node\n",
	    func);	
#endif
	return(-1);	
    }
    
    ret = xmlSecKeyWriteXml(key, type, cur);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write key node\n",
	    func);	
#endif
	xmlUnlinkNode(cur);
	xmlFreeNode(cur);
    }
    return(0);    
}

/**
 *
 *
 *
 *
 *
 */
static xmlSecKeyPtr 		
xmlSecRetrievalMethodRead(const xmlNodePtr retrievalMethodNode,
			 xmlSecTransformId id, xmlSecKeyType type,
			 int level, const xmlSecKeysReadContext *ctx) {
    static const char func[] = "xmlSecRetrievalMethodRead";			 
    xmlSecKeyPtr res = NULL;
    xmlNodePtr cur;
    xmlSecTransformStatePtr state = NULL;
    xmlDocPtr keyDoc = NULL;
    xmlChar *uri = NULL;
    int ret;
 
    if((retrievalMethodNode == NULL) || (ctx == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RetrievalMethod node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }    
    cur = xmlSecGetNextElementNode(retrievalMethodNode->children);
    
    /* read attributes first */
    uri = xmlGetProp(retrievalMethodNode, BAD_CAST "URI");
    if((uri == NULL) || (xmlStrlen(uri) == 0) || (uri[0] == '#')) {
	/* same document uri */
	if((ctx->allowedOrigins & xmlSecKeyOriginRetrievalDocument) == 0) {
	    /* not allowed */
	    xmlFree(uri);
	    return(NULL);
	}
    } else {
	/* remote document */
	if ((ctx->allowedOrigins & xmlSecKeyOriginRetrievalRemote) == 0) {
	    xmlFree(uri);
	    return(NULL);
	}
    }

    state = xmlSecTransformStateCreate(retrievalMethodNode->doc, NULL, (char*)uri);
    if(state == NULL){
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create transforms state\n",
	    func);
#endif	    
	goto done;
    }	

    /* first is optional Transforms node */
    if((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Transforms", xmlSecDSigNs)) {
	ret = xmlSecTransformsNodeRead(state, cur);
	if(ret < 0){
#ifdef DEBUG_XMLSEC    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Transforms\"\n",
		func);
#endif	    
	    goto done;
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }

    ret = xmlSecTransformStateFinal(state, xmlSecTransformResultBinary);
    if(ret < 0) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize transforms\n",
	    func);
#endif
	goto done;
    }

    if(cur != NULL) {
#ifdef DEBUG_XMLSEC
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node \"%s\" found\n",
	    func, cur->name);
#endif		
	goto done;
    }

    keyDoc = xmlRecoverMemory((char*)xmlBufferContent(state->curBuf),
                              xmlBufferLength(state->curBuf));
    if(keyDoc == NULL) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to parse keys doc\n");
#endif
	goto done;
    }
    
    res = xmlSecKeyNodesListRead(xmlDocGetRootElement(keyDoc), id, type, 
				level + 1, ctx);        
done:
    if(state != NULL) {
	xmlSecTransformStateDestroy(state);
    }
    if(uri != NULL) {
	xmlFree(uri);
    }
    if(keyDoc != NULL) {
	xmlFreeDoc(keyDoc);
    }
    return(res);
}


/**
 * xmlSecKeyReadEvpKey:
 * @pKey:
 * @id:
 * @origin:
 *
 * Converts pKey structure to xmlSecKey
 *
 */
xmlSecKeyPtr
xmlSecKeyReadEvpKey(EVP_PKEY *pKey, xmlSecTransformId id, xmlSecKeyOrigin origin) {
    static const char func[] = "xmlSecKeyReadEvpKey";
    xmlSecKeyPtr key = NULL;
    
    if(pKey == NULL) {
#ifdef DEBUG_XMLSEC    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: pKey is null\n",
	    func);
#endif	    
	return(NULL);
    }

    switch(pKey->type) {	
    case EVP_PKEY_RSA:
	if((id == xmlSecSignRsaSha1) || (id == xmlSecTransformUnknown)) {
	    key = xmlSecKeyCreate(xmlSecSignRsaKey, origin);
	    if(key != NULL) {
		xmlSecSignRsaKeySet(key, pKey->pkey.rsa);
		/* hack */
		pKey->pkey.rsa = NULL;
	    }
	}
	break;
    case EVP_PKEY_DSA:
	if((id == xmlSecSignDsaSha1) || (id == xmlSecTransformUnknown)) {
	    key = xmlSecKeyCreate(xmlSecSignDsaKey, origin);
	    if(key != NULL) {
		xmlSecSignDsaKeySet(key, pKey->pkey.dsa);
		/* hack */
		pKey->pkey.dsa = NULL;
	    }
	}
	break;
    default:	
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the key type %d is not supported\n",
	    func, pKey->type);	
#endif
	return(NULL);
    }
    if(key == NULL) {
#ifdef DEBUG_XMLSEC
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create xmlSecKey or key not found\n",
	    func);	
#endif
	return(NULL);
    }
    return(key);      
}




