/** 
 * XMLSec library
 *
 * Common XML Doc utility functions
 *
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/parser.h>
#include <libxml/xpathInternals.h>
#include <libxml/parserInternals.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>

/* 
 * xmlSecParseFile:
 * @filename:
 *
 * Loads XML Doc from file. We need a special version because of 
 * c14n issue. The code is copied from xmlSAXParseFileWithData() function.
 */
xmlDocPtr
xmlSecParseFile(const char *filename) {
    static const char func[] = "xmlSecParseFile";
    xmlDocPtr ret;
    xmlParserCtxtPtr ctxt;
    char *directory = NULL;
    
    if(filename == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: filename is null\n", 
	    func);	
#endif
	return(NULL);	
    }


    xmlInitParser();

    ctxt = xmlCreateFileParserCtxt(filename);
    if (ctxt == NULL) {
	return(NULL);
    }

    /* todo: set directories from current doc? */    
    if ((ctxt->directory == NULL) && (directory == NULL))
        directory = xmlParserGetDirectory(filename);
    if ((ctxt->directory == NULL) && (directory != NULL))
        ctxt->directory = (char *) xmlStrdup((xmlChar *) directory);

    /* required for c14n! */
    ctxt->loadsubset = XML_DETECT_IDS | XML_COMPLETE_ATTRS; 
    ctxt->replaceEntities = 1;
    
    xmlParseDocument(ctxt);

    if(ctxt->wellFormed) { 
	ret = ctxt->myDoc;
    } else {
       ret = NULL;
       xmlFreeDoc(ctxt->myDoc);
       ctxt->myDoc = NULL;
    }
    xmlFreeParserCtxt(ctxt);    
    return(ret);
    
}

/* 
 * xmlSecParseMemory:
 * @filename:
 *
 * Loads XML Doc from memory. We need a special version because of 
 * c14n issue. The code is copied from xmlSAXParseMemory() function.
 */
xmlDocPtr
xmlSecParseMemory(const char *buffer, int size, int recovery) {
    static const char func[] = "xmlSecParseMemory";
    xmlDocPtr ret;
    xmlParserCtxtPtr ctxt;
    
    if(buffer == NULL){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: buffer is null\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    ctxt = xmlCreateMemoryParserCtxt(buffer, size);
    if (ctxt == NULL) return(NULL);

    /* required for c14n! */
    ctxt->loadsubset = XML_DETECT_IDS | XML_COMPLETE_ATTRS; 
    ctxt->replaceEntities = 1;

    xmlParseDocument(ctxt);

    if((ctxt->wellFormed) || recovery) {
	ret = ctxt->myDoc; 
    } else {
       ret = NULL;
       xmlFreeDoc(ctxt->myDoc);
       ctxt->myDoc = NULL;
    }
    xmlFreeParserCtxt(ctxt);    
    return(ret);
}

/**
 * xmlSecFindChild:
 * @parent:
 * @name:
 * @ns:
 *
 * Lookups a direct child of the node having given name and namespace href
 * 
 */
xmlNodePtr
xmlSecFindChild(const xmlNodePtr parent, const xmlChar *name, const xmlChar *ns) {
    static const char func[] = "xmlSecFindChild";
    xmlNodePtr cur;
        
    if((parent == NULL) || (name == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the node or name is null\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    cur = parent->children;
    while(cur != NULL) {
        if(cur->type == XML_ELEMENT_NODE) {
	    if(xmlSecCheckNodeName(cur, name, ns)) {
		return(cur);
	    }
	}
	cur = cur->next;
    }
    return(NULL);
}

/** 
 * xmlSecFindNodeById:
 * @parent:
 * @id:
 *
 * Lookups the node with attribute "Id" equal to given one in the given 
 * node subtree 
 */
xmlNodePtr
xmlSecFindNodeById(const xmlNodePtr parent, const xmlChar *id) {
    static const char func[] = "xmlSecFindNodeById";
    xmlNodePtr cur;
    
    if((parent == NULL) || (id == NULL)){
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the node or id is null\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    cur = parent;
    while(cur != NULL) {
        if(cur->type == XML_ELEMENT_NODE) {	    
	    xmlChar* attr;
	    xmlNodePtr ret;
	    
	    attr = xmlGetProp(cur, BAD_CAST "Id");
	    if(xmlStrEqual(id, attr)) {
		xmlFree(attr);
		return(cur);
	    }
	    xmlFree(attr);
	    
	    if(cur->children != NULL) {
	        ret = xmlSecFindNodeById(cur->children, id);
		if(ret != NULL) {
	    	    return(ret);	    
		}
	    }
	}
	cur = cur->next;
    }
    return(NULL);
}

/**
 * xmlSecGetChildNodeSet:
 * @parent:
 * @nodeSet:
 * @withComments
 *
 * Creates a node set that includes all nodes in given subtree
 *
 */
xmlNodeSetPtr
xmlSecGetChildNodeSet(const xmlNodePtr parent, xmlNodeSetPtr nodeSet, int withComments) {
    static const char func[] = "xmlSecGetChildNodeSet";
    xmlNodePtr cur;
    xmlAttrPtr attr;
    
    if(parent == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the node  is null\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    /* add the node */
    if(nodeSet == NULL) {	
	nodeSet = xmlXPathNodeSetCreate(NULL);
	if(nodeSet == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create node set\n",
		func);
#endif
	    return(NULL);
	}	
    } 

    switch(parent->type) {
    case XML_COMMENT_NODE:
        if(!withComments) return(nodeSet);
        xmlXPathNodeSetAddUnique(nodeSet, parent);
        return(nodeSet);	
    case XML_ELEMENT_NODE:
        xmlXPathNodeSetAddUnique(nodeSet, parent);
        break;
    case XML_TEXT_NODE:
        xmlXPathNodeSetAddUnique(nodeSet, parent);
        return(nodeSet);
    case XML_PI_NODE:
        xmlXPathNodeSetAddUnique(nodeSet, parent);
        return(nodeSet);
    default:
        return(nodeSet);
    }
    
    /* add all attrs */
    attr = parent->properties; 
    while (attr != NULL) {
        xmlXPathNodeSetAddUnique(nodeSet, (xmlNodePtr)attr); 
        attr = attr->next; 
    }	
    cur = parent->children;
    
    /* add all childrens */
    while(cur != NULL) {
	nodeSet = xmlSecGetChildNodeSet(cur, nodeSet, withComments);
	if(nodeSet == NULL) {
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add child namespace\n",
		func);	
#endif
	    return(NULL);	    
	}
	cur = cur->next;
    }
    
    return(nodeSet);
    
}

/** 
 * xmlSecCheckNodeName:
 * @cur:
 * @name:
 * @ns:
 *
 * Checks that the node has a given name and a given namespace href
 */
int
xmlSecCheckNodeName(const xmlNodePtr cur, const xmlChar *name, const xmlChar *ns) {
    static const char func[] = "xmlSecCheckNodeName";
    
    if((cur != NULL) && xmlStrEqual(cur->name, name)) {
	if(cur->ns == NULL && ns == NULL) {
	    return(1);
	} else if(cur->ns == NULL) {
	    xmlNsPtr tmp;

	    tmp = xmlSearchNs(cur->doc, cur, NULL);
	    if(tmp != NULL && xmlStrEqual(tmp->href, ns)) {
		return(1);
	    }
	} else if(xmlStrEqual(cur->ns->href, ns)){
	    return(1);	
	}
    } 
    return(0);   
}

/**
 * xmlSecAddChild
 * @parent:
 * @name:
 * @ns:
 *
 */
xmlNodePtr		
xmlSecAddChild(xmlNodePtr parent, const xmlChar *name, const xmlChar *ns) {
    static const char func[] = "xmlSecAddChild";
    xmlNodePtr cur;
    xmlNodePtr text;
        
    if((parent == NULL) || (name == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: parent node or name is null\n",
	    func);	
#endif
	return(NULL);
    }

    if(parent->children == NULL) {
        /* TODO: add indents */
	text = xmlNewText(BAD_CAST "\n"); 
        if(text == NULL) {	
#ifdef DEBUG_XMLSEC
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create line break text node\n",
	        func);	
#endif
	    return(NULL);
	}
	xmlAddChild(parent, text);
    }

    /* TODO: add namespaces support */
    cur = xmlNewChild(parent, NULL, name, NULL);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create node\n",
	    func);	
#endif
	return(NULL);
    }
    if(ns != NULL) {
        /* TODO: add namespaces support */
    }

    /* TODO: add indents */
    text = xmlNewText(BAD_CAST "\n"); 
    if(text == NULL) {	
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create line break text node\n",
	    func);	
#endif
	return(NULL);
    }
    xmlAddChild(parent, text);

    return(cur);
}

/**
 * xmlSecAddNextSibling
 * @parent:
 * @name:
 * @ns:
 *
 */
xmlNodePtr
xmlSecAddNextSibling(xmlNodePtr node, const xmlChar *name, const xmlChar *ns) {
    static const char func[] = "xmlSecAddNextSibling";
    xmlNodePtr cur;
    xmlNodePtr text;
    
    if((node == NULL) || (name == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: sibling node or name is null\n",
	    func);	
#endif
	return(NULL);
    }

    cur = xmlNewNode(NULL, name);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create node\n",
	    func);	
#endif
	return(NULL);
    }
    if(ns != NULL) {
        /* TODO: add namespaces support */
    }
    xmlAddNextSibling(node, cur);

    /* TODO: add indents */
    text = xmlNewText(BAD_CAST "\n");
    if(text == NULL) {	
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create line break text node\n",
	    func);	
#endif
	return(NULL);
    }
    xmlAddNextSibling(node, text);
    
    return(cur);
}

/**
 * xmlSecAddPrevSibling
 * @parent:
 * @name:
 * @ns:
 *
 */
xmlNodePtr
xmlSecAddPrevSibling(xmlNodePtr node, const xmlChar *name, const xmlChar *ns) {
    static const char func[] = "xmlSecAddPrevSibling";
    xmlNodePtr cur;
    xmlNodePtr text;
    
    if((node == NULL) || (name == NULL)) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: sibling node or name is null\n",
	    func);	
#endif
	return(NULL);
    }

    cur = xmlNewNode(NULL, name);
    if(cur == NULL) {
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create node\n",
	    func);	
#endif
	return(NULL);
    }
    if(ns != NULL) {
        /* TODO: add namespaces support */
    }
    xmlAddPrevSibling(node, cur);

    /* TODO: add indents */
    text = xmlNewText(BAD_CAST "\n");
    if(text == NULL) {	
#ifdef DEBUG_XMLSEC
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create line break text node\n",
	    func);	
#endif
	return(NULL);
    }
    xmlAddPrevSibling(node, text);

    
    return(cur);
}

/**
 * xmlSecGetNextElementNode
 * @cur:
 *
 *
 */
xmlNodePtr
xmlSecGetNextElementNode(xmlNodePtr cur) {
    
    while((cur != NULL) && (cur->type != XML_ELEMENT_NODE)) {
	cur = cur->next;
    }
    return(cur);
}



