/** 
 * XML Security standards test: XMLDSig
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>
#include <time.h>


#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysmngr.h>
#include <xmlsec/hmac.h>
#include <xmlsec/x509.h>
#include <xmlsec/x509mngr.h>
#include <xmlsec/xmldsig.h>


void testDSigUsage(FILE *f, const char *command);

void testDSigUsage(FILE *f, const char *command) {
    fprintf(stderr, "Usage:\n");
    fprintf(f,   "%s sign|verify [<options>] [<filename> [<filename> ... ]] \n", command);
    fprintf(f, "where <options> are\n");
    fprintf(f, "== Key Manager:\n");
    fprintf(f, "  --privkey <file>[,<cafile1>[,<cafile2>[...]]] the private PEM key file\n");
    fprintf(f, "                        and the list of cert file that verify it\n");
    fprintf(f, "  --pubkey <file>       the public PEM key file\n");
    fprintf(f, "  --hmackey <file>      the HMAC key file\n");
    fprintf(f, "  --keys-origins <list> the list of comma separated allowed key origins\n");
    fprintf(f, "                        from the following set: \n");
    fprintf(f, "                           \"keymanager\", \"keyname\", \"keyvalue\",\n");
    fprintf(f, "                           \"retrieval-doc\", \"retrieval-remote\",\n");
    fprintf(f, "                           \"x509\", \"pgp\"\n");
    fprintf(f, "                        by default, all origins are allowed\n");
    fprintf(f, "  --keys <keys>         load keys from file\n");
    fprintf(f, "== Result generation/printing:\n");
    fprintf(f, "  --dump                dump the debug verification results\n");
    fprintf(f, "  --ignore-manifests    don't process Manifests\n");
    fprintf(f, "  --store-sign-refs     store signed references before digesting\n" );
    fprintf(f, "  --store-manifest-refs store Manifests references before digesting\n"); 
    fprintf(f, "  --store-signatures    store SignedInfo before signing\n");
    fprintf(f, "  --store-all           store everything (combination of \"--store-sign-refs\",\n");
    fprintf(f, "                        \"--store-manifest-refs\" and \"--store-signatures\")\n");    fprintf(f, "  --CAfile <file>       the x509 root ca file\n");
    fprintf(f, "== Misc.:\n");
    fprintf(f, "  --retry <number>      repeat operation with each file <number> times\n" );
}

/**
 * Init/Shutdown
 */
int			testDSigInit			(int sign);
void			testDSigShutdown		(void);

/**
 * Parsing the command line
 */
int			testDSigReadKeyOrigins		(char *keyOrigins);
int			testDSigReadKey			(int privateKey, 
							 char *param);
int			testDSigReadHmacKey		(char *filename);							 
int 			testDSigReadKeysFile		(char *keysFile);
int			testDSigReadRetries		(char *retries);
int 			testDSigReadCA			(char *cafile);







typedef struct _xmlDSigResult {
    size_t			signaturesOk;
    size_t			signaturesFail;
    size_t			signRefOk;
    size_t			signRefFail;
    size_t			manifestRefOk;
    size_t			manifestRefFail;
} xmlDSigResult, *xmlDSigResultPtr;


void			testDSigGetResult	(xmlDSigSignaturePtr signature, 
						 xmlDSigResultPtr result, 
						 int direction);
int 			testDSig		(const char* filename, 
						 int generate);

int 			testDSigGenerate	(xmlDocPtr doc);
int 			testDSigValidate	(xmlDocPtr doc);


xmlDSigCtxPtr dsigCtx = NULL;
xmlSecSimpleKeyMngrPtr keyMgr = NULL; 
xmlSecSimpleX509MngrPtr x509Mngr = NULL;
xmlSecKeysReadContext keysReadCtx;
int retryNumber = 1;

clock_t total_time = 0;
int doDump = 0;

int main(int argc, char **argv) {
    int sign = 0;    
    int ret = -1;
    int res = -1;
    int pos;
    int argsOk = 0;    

    /*
     * Parse command line
     */
    if( argc < 3 ) {
	fprintf(stderr, "Error: wrong number of arguments.\n");
	goto done;		
    } 
    if(strcmp(argv[1], "sign") == 0) {
	sign = 1;
    } else if(strcmp(argv[1], "verify") == 0) {
	sign = 0;
    } else {
        fprintf(stderr, "Error: bad mode \"%s\".\n", argv[1]);
        goto done;
    }
    
    ret = testDSigInit(sign);
    if(ret < 0) {
	goto done;
    }
    
    /* load options */
    pos = 2;
    while(pos < argc) {
	if(argv[pos][0] != '-') {
	    /* it's filename? */
	    break;
	} else if((strcmp(argv[pos], "--keys") == 0) && (pos + 1 < argc)) {	  
	    ret = testDSigReadKeysFile(argv[++pos]);
	} else if((strcmp(argv[pos], "--retry") == 0) && (pos + 1 < argc)) {
	    ret = testDSigReadRetries(argv[++pos]);
	} else if((strcmp(argv[pos], "--CAfile") == 0) && (pos + 1 < argc)) {
	    ret = testDSigReadCA(argv[++pos]); 
	} else if((strcmp(argv[pos], "--privkey") == 0) && (pos + 1 < argc)) {
	    ret = testDSigReadKey(1, argv[++pos]); 
	} else if((strcmp(argv[pos], "--pubkey") == 0) && (pos + 1 < argc)) {
	    ret = testDSigReadKey(0, argv[++pos]); 
	} else if((strcmp(argv[pos], "--hmackey") == 0) && (pos + 1 < argc)) {
	    ret = testDSigReadHmacKey(argv[++pos]); 
	} else if((strcmp(argv[pos], "--keys-origins") == 0) && (pos + 1 < argc)){
	    ret = testDSigReadKeyOrigins(argv[++pos]);
	} else if(strcmp(argv[pos], "--dump") == 0) {
	    doDump = 1;
	} else if(strcmp(argv[pos], "--ignore-manifests") == 0) {
	    dsigCtx->processManifests = 0;
	} else if(strcmp(argv[pos], "--store-sign-refs") == 0) {
	    dsigCtx->storeReferences = 1;
	} else if(strcmp(argv[pos], "--store-manifest-refs") == 0) {
	    dsigCtx->storeManifests = 1;
	} else if(strcmp(argv[pos], "--store-signatures") == 0) {
	    dsigCtx->storeSignatures = 1;
	} else if(strcmp(argv[pos], "--store-all") == 0) {
	    dsigCtx->storeReferences = 1;
	    dsigCtx->storeManifests = 1;
	    dsigCtx->storeSignatures = 1;
	} else {
	    fprintf(stderr, "Error: bad option \"%s\".\n", argv[pos]);
	    goto done;
	}
	if(ret < 0) {
	    goto done;
	}
	++pos;
    }
    argsOk = 1;

    while(pos < argc) { 
	if(sign) {
	    res = testDSig(argv[pos++], 1);	    
	    if(res < 0) {	
		break;
	    }
	} else {
	    res = testDSig(argv[pos++], 0);
	    if(res < 0) {
		break;
	    }
	}
    }
    fprintf(stderr, "Executed %d tests in %ld msec\n", retryNumber, total_time / (CLOCKS_PER_SEC / 1000));

done:
    if(!argsOk) {
	testDSigUsage(stderr, argv[0]);
    }
        
    testDSigShutdown();
    return((res >= 0) ? 0 : 1);
}

int 
testDSig(const char* filename, int generate) {
    xmlDocPtr doc;
    int ret = 0;
    int i;
    clock_t start_time;
    
    /*
     * build an XML tree from a the file; we need to add default
     * attributes and resolve all character and entities references
     */
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
    xmlSubstituteEntitiesDefault(1);

    doc = xmlParseFile(filename);
    if (doc == NULL) {
	fprintf(stderr, "Error: unable to parse file \"%s\"\n", filename);
	return(-1);
    }
    
    /*
     * Check the document is of the right kind
     */    
    if(xmlDocGetRootElement(doc) == NULL) {
        fprintf(stderr,"Error: empty document for file \"%s\"\n", filename);
	xmlFreeDoc(doc);
	return(-1);
    }

    start_time = clock();
    for(i = 0; i < retryNumber; ++i) {
        /*
	 * Generate/verify signature
	 */      
	if(generate) {
	    fprintf(stderr, "Generating xml signature for \"%s\"\n", filename);	
	    ret = testDSigGenerate(doc);
	} else {
	    fprintf(stderr, "Validating xml signature in \"%s\"\n", filename);	
	    ret = testDSigValidate(doc);
	}
    }    
    
    total_time += clock() - start_time;
           
    /*
     * Cleanup
     */ 
    xmlFreeDoc(doc);    

    return(ret);
}
	
int testDSigGenerate(xmlDocPtr doc) {    
    xmlDSigSignaturePtr signature = NULL;
    xmlChar *result = NULL;	
    int len;
    int ret;
    int res = -1;

    ret = xmlDSigGenerate(dsigCtx, xmlDocGetRootElement(doc), &signature);
    if(ret < 0) {
        fprintf(stderr,"Error: xmlDSigGenerate() failed \n");
	goto done;    
    } 
    	
    /*
     * Print document out in default UTF-8 encoding
     */
    xmlDocDumpMemoryEnc(doc, &result, &len, NULL);
    if(result == NULL) {
        fprintf(stderr,"Error: failed to dump document to memory\n");
	goto done;
    }
    fwrite(result, len, 1, stdout);
    res = 0;
    
done:    
    if(result != NULL) {
	xmlFree(result);        
    }
    if(signature != NULL) {
	xmlDSigSignatureDestroyAll(signature);
    }
    return(res);
}

int testDSigValidate(xmlDocPtr doc) {    
    xmlDSigSignaturePtr signature = NULL;
    xmlDSigResult result;
    int ret;
    int res = -1;
    	    
    ret = xmlDSigValidate(dsigCtx, xmlDocGetRootElement(doc), &signature);
    if((ret < 0) || (signature == NULL)){
	fprintf(stdout,"ERROR\n");
	goto done;
    } 
	    
    if(doDump) {
    	xmlDSigDebugSignatureDumpAll(stderr, signature);
    }	
	    
    /** 
     * we will simply walk thru and calculate the number of 
     * ok/fails
     */
    memset(&result, 0, sizeof(result));
    testDSigGetResult(signature, &result, 0);
	
    /* 
     * if one signature fails then we are failed. 
     * in real application the rules could be different
     */
    fprintf(stderr, "= Status:\n");
    fprintf(stderr, "== Signatures ok: %d\n", result.signaturesOk);
    fprintf(stderr, "== Signatures fail: %d\n", result.signaturesFail);
    fprintf(stderr, "== SignedInfo Ref ok: %d\n", result.signRefOk);
    fprintf(stderr, "== SignedInfo Ref fail: %d\n", result.signRefFail);
    fprintf(stderr, "== Manifest Ref ok: %d\n", result.manifestRefOk);
    fprintf(stderr, "== Manifest Ref fail: %d\n", result.manifestRefFail);

    if(result.signaturesFail == 0) {
	fprintf(stdout, "OK\n");  
	res = 0;    
    }else {
	fprintf(stdout, "FAIL\n");
	res = -1;    
    }
    	    	
done:
    if(signature != NULL) {
	xmlDSigSignatureDestroyAll(signature);
    }
    return(res);
}

void
testDSigGetResult(xmlDSigSignaturePtr signature, xmlDSigResultPtr result, int direction) {
    xmlDSigReferencePtr ref;
    
    if((signature == NULL) || (result == NULL)) {
	fprintf(stderr, "Error: signature or result is null\n");
	return;
    }
    
    if(signature->result == xmlSecTransformStatusOk) {
	++(result->signaturesOk);
    } else {
	++(result->signaturesFail);
    }
    
    ref = signature->firstSignRef;
    while(ref != NULL) {
	if(ref->result == xmlSecTransformStatusOk) {
	    ++(result->signRefOk);
	} else {
	    ++(result->signRefFail);
	}
	ref = ref->next;
    }

    ref = signature->firstManifestRef;
    while(ref != NULL) {
	if(ref->result == xmlSecTransformStatusOk) {
	    ++(result->manifestRefOk);
	} else {
	    ++(result->manifestRefFail);
	}
	ref = ref->next;
    }
    
    if((direction <= 0) && (signature->prev != NULL)) {
	testDSigGetResult(signature->prev, result, -1);
    }

    if((direction >= 0) && (signature->next != NULL)) {
	testDSigGetResult(signature->next, result, 1);
    }
    
}

/**
 * Parsing the command line
 */
int			
testDSigReadKeyOrigins(char *keyOrigins) {
    xmlSecKeyOrigin res = xmlSecKeyOriginDefault;
    char *p;
    
    p = strtok(keyOrigins, ",");
    while(p != NULL) {
	if(strcmp(p, "keymanager") == 0) {
	    res |= xmlSecKeyOriginKeyManager;
	} else if(strcmp(p, "keyname") == 0) {
	    res |= xmlSecKeyOriginKeyManager | xmlSecKeyOriginKeyName;
	} else if(strcmp(p, "keyvalue") == 0) {
	    res |= xmlSecKeyOriginKeyValue;
	} else if(strcmp(p, "retrieval-doc") == 0) {
	    res |= xmlSecKeyOriginRetrievalDocument;
	} else if(strcmp(p, "retrieval-remote") == 0) {
	    res |= xmlSecKeyOriginRetrievalRemote;
	} else if(strcmp(p, "x509") == 0) {
	    res |= xmlSecKeyOriginX509;
	} else if(strcmp(p, "pgp") == 0) {
	    res |= xmlSecKeyOriginPGP;
	} else {
	    fprintf(stderr, "Error: unknown key origin: \"%s\" (ignored)\n", p);
	    return(-1);
	}
	p = strtok(NULL, ",");
    }    
    keysReadCtx.allowedOrigins = res;
    return(0);
}

int
testDSigReadKey	(int privateKey, char *param) {
    char *p;
    xmlSecKeyPtr key;
    int ret;
    
    p = strtok(param, ","); 
    if(privateKey) {
	key = xmlSecSimpleKeyMngrLoadPrivateKey(keyMgr, p, NULL, NULL);
    } else {
	key = xmlSecSimpleKeyMngrLoadPublicKey(keyMgr, p, NULL, NULL);
    }
    if(key == NULL) {
	fprintf(stderr, "Error: failed to load key from \"%s\"\n", p);
	return(-1);
    }
    
    p = strtok(NULL, ",");
    while((p != NULL) && (privateKey)) {
	ret = xmlSecSimpleKeyMngrAddCertToKey(key, p);
	if(ret < 0){
	    fprintf(stderr, "Error: failed to load cert from \"%s\"\n", p);
	    return(-1);
	}
	p = strtok(NULL, ","); 
    }
    /* TODO: */
    return(0);
}

int
testDSigReadHmacKey(char *filename) {
    FILE *f;
    unsigned char buf[1024];
    xmlSecMacHmacKeyData hmacData;
    xmlSecKeyPtr key;
    int ret;
    
    
    f = fopen(filename, "r");
    if(f == NULL) {
	fprintf(stderr, "Error: failed to open file \"%s\" \n", filename);
	return(-1);
    }
    
    ret = fread(buf, 1, sizeof(buf), f);
    if(ret < 0) {
	fprintf(stderr, "Error: failed to read file \"%s\" \n", filename);
	fclose(f);
	return(-1);
    }
    fclose(f);    
    hmacData.key = buf;
    hmacData.size = ret;
    
    /* HMAC */    
    key = xmlSecKeyGenerate(xmlSecMacHmacKey, &hmacData);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create hmac key\n"); 
	return(-1);
    }    
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add hmac key\n"); 
	return(-1);
    }
    return(0);
}

int
testDSigReadKeysFile(char *keysFile) {
    int ret;
    
    ret = xmlSecSimpleKeyMngrLoad(keyMgr, keysFile);
    if(ret < 0) {
	fprintf(stderr, "Error: failed to load keys from \"%s\".\n", keysFile);
	return(-1);
    }
    return(0);
}

int
testDSigReadRetries(char *retries) {
    
    if(sscanf(retries, "%d", &retryNumber) != 1) {
	fprintf(stderr, "Error: bad number \"%s\".\n", retries);
	return(-1);
    }	    
    return(0);
}

int
testDSigReadCA(char *cafile) {
    int ret;
	    
    ret = xmlSecSimpleX509MngrAddTrustedCA(x509Mngr, cafile); 
    if(ret < 0) {
	fprintf(stderr, "Error: CAfile \"%s\" load failed.\n", cafile);
	return(-1);
    }     
    return(0);
}

/**
 * Init/Shutdown
 */
int 
testDSigInit(int sign) {
    int rnd_seed = 0;

    /* 
     * Init OpenSSL
     */
    OpenSSL_add_all_algorithms();
    ERR_load_crypto_strings(); 
    while (RAND_status() != 1) {
	RAND_seed(&rnd_seed, sizeof(rnd_seed));
    }

    /*
     * Init libxml
     */     
    xmlInitParser();
    LIBXML_TEST_VERSION

    /*
     * Init xmlsec
     */
    xmlSecInit();    


    /** 
     * Create Keys and x509 managers
     */
    keyMgr = xmlSecSimpleKeyMngrCreate();    
    if(keyMgr == NULL) {
	fprintf(stderr, "Error: failed to create keys manager\n");
	return(-1);
    }
    x509Mngr = xmlSecSimpleX509MngrCreate(sign);
    if(x509Mngr == NULL) {
	fprintf(stderr, "Error: failed to create x509 manager\n");
	return(-1);
    }
    
    memset(&keysReadCtx, 0, sizeof(keysReadCtx));

    keysReadCtx.allowedOrigins = xmlSecKeyOriginAll; /* by default all keys are accepted */
    keysReadCtx.maxRetrievals = 1; /* one level retrievals only */
    keysReadCtx.findKeyCallback = xmlSecSimpleKeyMngrFindKey;
    keysReadCtx.findKeyContext = keyMgr;
    keysReadCtx.x509Context = x509Mngr; 
    keysReadCtx.x509VerifyCallback = (xmlSecX509VerifyCallback)xmlSecSimpleX509MngrVerify;
    keysReadCtx.x509AddCRLCallback = (xmlSecX509AddCRLCallback)xmlSecSimpleX509MngrAddCRL;

    /**
     * Init DSig context
     */    
    dsigCtx = xmlDSigCtxCreate(&keysReadCtx);
    if(dsigCtx == NULL) {
        fprintf(stderr,"Error: failed to create DSig context\n");
	return(-1);
    }
        
    /**
     * Set default values to process manifests and store nothing
     * Overwrite this options thru command line if needed!
     */
    dsigCtx->processManifests = 1;
    dsigCtx->storeSignatures = 0;
    dsigCtx->storeReferences = 0;
    dsigCtx->storeManifests = 0;
    return(0);    
}

void
testDSigShutdown(void) {
    /* destroy xmlsec objects */
    if(x509Mngr != NULL) {
	xmlSecSimpleX509MngrDestroy(x509Mngr);
    }
    if(keyMgr != NULL) {
	xmlSecSimpleKeyMngrDestroy(keyMgr);
    }
    if(dsigCtx != NULL) {
	xmlDSigCtxDestroy(dsigCtx);
    }

    xmlSecShutdown();
            
    /* 
     * Shutdown libxml
     */
    xmlCleanupParser();
    xmlMemoryDump();

    /**
     * Shutdown OpenSSL
     */    
    RAND_cleanup();
    ERR_free_strings();
    EVP_cleanup();
    X509_TRUST_cleanup();
}

